# -*- coding: utf-8 -*-

import base64
import json
from typing import Dict
from PIL import Image
from io import BytesIO

from uiviewer._logger import logger


def file2base64(path: str) -> str:
    with open(path, "rb") as file:
        base64_encoded = base64.b64encode(file.read())
        return base64_encoded.decode('utf-8')


def image2base64(image: Image.Image, format: str = "PNG") -> str:
    """
    PIL Image to base64 string
    """
    buffered = BytesIO()
    image.save(buffered, format=format)
    return base64.b64encode(buffered.getvalue()).decode('utf-8')


def str2json(s: str) -> Dict:
    try:
        json_obj = json.loads(s)
        return json_obj
    except json.JSONDecodeError as e:
        logger.error(f"Invalid JSON data: {e}")
        return {}


# -*- coding: utf-8 -*-

import abc
import tempfile
from typing import List, Dict, Union, Tuple, Optional
from functools import cached_property  # python3.8+

from PIL import Image
from requests import request
import tidevice
import adbutils
import wda
import uiautomator2 as u2
from hmdriver2 import hdc
from fastapi import HTTPException

from uiviewer._utils import file2base64, image2base64
from uiviewer._models import Platform, BaseHierarchy
from uiviewer.parser import android_hierarchy, ios_hierarchy, harmony_hierarchy


def list_serials(platform: str) -> List[str]:
    devices = []
    if platform == Platform.ANDROID:
        raws = adbutils.AdbClient().device_list()
        devices = [item.serial for item in raws]
    elif platform == Platform.IOS:
        raw = tidevice.Usbmux().device_list()
        devices = [d.udid for d in raw]
    else:
        devices = hdc.list_devices()

    return devices


class DeviceMeta(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def take_screenshot(self) -> str:
        pass

    def dump_hierarchy(self) -> Dict:
        pass


class HarmonyDevice(DeviceMeta):
    def __init__(self, serial: str):
        self.serial = serial
        self.hdc = hdc.HdcWrapper(serial)

    @cached_property
    def _display_size(self) -> Tuple:
        return self.hdc.display_size()

    def take_screenshot(self) -> str:
        with tempfile.NamedTemporaryFile(delete=True, suffix=".png") as f:
            path = f.name
            self.hdc.screenshot(path)
            return file2base64(path)

    def dump_hierarchy(self) -> BaseHierarchy:
        packageName, pageName = self.hdc.current_app()
        raw: Dict = self.hdc.dump_hierarchy()
        hierarchy: Dict = harmony_hierarchy.convert_harmony_hierarchy(raw)
        return BaseHierarchy(
            jsonHierarchy=hierarchy,
            activityName=pageName,
            packageName=packageName,
            windowSize=self._display_size,
            scale=1
        )


class AndroidDevice(DeviceMeta):
    def __init__(self, serial: str):
        self.serial = serial
        self.d: u2.Device = u2.connect(serial)

    @cached_property
    def _window_size(self) -> Tuple:
        return self.d.window_size()

    def take_screenshot(self) -> str:
        img: Image.Image = self.d.screenshot()
        return image2base64(img)

    def dump_hierarchy(self) -> BaseHierarchy:
        current = self.d.app_current()
        page_xml = self.d.dump_hierarchy()
        page_json = android_hierarchy.convert_android_hierarchy(page_xml)
        return BaseHierarchy(
            jsonHierarchy=page_json,
            activityName=current['activity'],
            packageName=current['package'],
            windowSize=self._window_size,
            scale=1
        )


class IosDevice(DeviceMeta):
    def __init__(self, udid: str, wda_url: str, max_depth: int) -> None:
        self.udid = udid
        self.wda_url = wda_url
        self._max_depth = max_depth
        self.client = wda.Client(wda_url)

    @property
    def max_depth(self) -> int:
        return int(self._max_depth) if self._max_depth else 30

    @cached_property
    def scale(self) -> int:
        return self.client.scale

    @cached_property
    def _window_size(self) -> Tuple:
        return self.client.window_size()

    def _check_wda_health(self) -> bool:
        resp = request("GET", f"{self.wda_url}/status", timeout=5).json()
        state = resp.get("value", {}).get("state")
        return state == "success"

    def take_screenshot(self) -> str:
        img: Image.Image = self.client.screenshot()
        return image2base64(img)

    def _current_bundle_id(self) -> str:
        resp = request("GET", f"{self.wda_url}/wda/activeAppInfo", timeout=10).json()
        bundleId = resp.get("value", {}).get("bundleId", None)
        return bundleId

    def dump_hierarchy(self) -> BaseHierarchy:
        self.client.appium_settings({"snapshotMaxDepth": self.max_depth})
        data: Dict = self.client.source(format="json")
        hierarchy: Dict = ios_hierarchy.convert_ios_hierarchy(data, self.scale)
        return BaseHierarchy(
            jsonHierarchy=hierarchy,
            activityName=None,
            packageName=self._current_bundle_id(),
            windowSize=self._window_size,
            scale=self.scale
        )


def get_device(platform: str, serial: str, wda_url: str, max_depth: int) -> Union[HarmonyDevice, AndroidDevice, IosDevice]:
    if platform == Platform.HARMONY:
        return HarmonyDevice(serial)
    elif platform == Platform.ANDROID:
        return AndroidDevice(serial)
    else:
        return IosDevice(serial, wda_url, max_depth)


# Global cache for devices
cached_devices = {}


def init_device(platform: str, serial: str, wda_url: str, max_depth: int):

    if serial not in list_serials(platform):
        raise HTTPException(status_code=500, detail=f"Device<{serial}> not found")

    try:
        device: Union[HarmonyDevice, AndroidDevice] = get_device(platform, serial, wda_url, max_depth)
        cached_devices[(platform, serial)] = device

        if isinstance(device, IosDevice):
            return device._check_wda_health()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

    return True


# -*- coding: utf-8 -*-

import os
import webbrowser
import uvicorn
import threading

from fastapi import FastAPI, Request, HTTPException
from fastapi.staticfiles import StaticFiles
from fastapi.responses import JSONResponse

from uiviewer.routers import api
from uiviewer._models import ApiResponse


app = FastAPI()


current_dir = os.path.dirname(os.path.abspath(__file__))
static_dir = os.path.join(current_dir, "static")

app.mount("/static", StaticFiles(directory=static_dir), name="static")

app.include_router(api.router)


@app.exception_handler(Exception)
def global_exception_handler(request: Request, exc: Exception):
    return JSONResponse(
        status_code=500,
        content=ApiResponse(success=False, message=str(exc)).dict()
    )


@app.exception_handler(HTTPException)
def http_exception_handler(request: Request, exc: HTTPException):
    return JSONResponse(
        status_code=exc.status_code,
        content=ApiResponse(success=False, message=exc.detail).dict(),
    )


def open_browser(port):
    webbrowser.open_new(f"http://127.0.0.1:{port}")


def run(port=8000):
    timer = threading.Timer(1.0, open_browser, args=[port])
    timer.daemon = True
    timer.start()

    uvicorn.run(app, host="127.0.0.1", port=port)


if __name__ == "__main__":
    run()


# -*- coding: utf-8 -*-

from typing import Dict


class XPathLiteGenerator:
    def __init__(self, platform: str, treedata: Dict):
        """
        Initializes the XPathLiteGenerator class.

        Args:
        platform (str): The platform type (e.g., 'ios', 'android').
        treedata (Dict): The JSON tree structure data.

        Returns:
        None
        """
        self.platform = platform
        self.treedata = treedata
        self.node_map = self._build_node_map(treedata)

    def _build_node_map(self, node: Dict, node_map: Dict[str, Dict] = None) -> Dict[str, Dict]:
        """
        Builds a node map for quick lookup by node _id.

        Args:
        node (Dict): The current node.
        node_map (Dict[str, Dict], optional): The node map. Default is an empty dictionary.

        Returns:
        Dict[str, Dict]: The node map.
        """
        if node_map is None:
            node_map = {}
        node_map[node["_id"]] = node
        if "children" in node:
            for child in node["children"]:
                self._build_node_map(child, node_map)
        return node_map

    def _find_node_by_id(self, target_id: str) -> Dict:
        """
        Finds a node by its ID.

        Args:
        target_id (str): The target node ID.

        Returns:
        Dict: The target node.
        """
        return self.node_map.get(target_id)

    def _get_value(self, node: Dict) -> str:
        """
        Gets the specific attribute value of a node to generate part of the XPath expression.

        Args:
        node (Dict): The current node.

        Returns:
        str: Part of the XPath expression.
        """
        if node.get("resourceId"):
            return f'//*[@resource-id="{node["resourceId"]}"]'
        elif node.get("text"):
            return f'//*[@text="{node["text"]}"]'
        elif node.get("description"):
            return f'//*[@content-desc="{node["description"]}"]'
        elif node.get("label"):
            return f'//*[@label="{node["label"]}"]'
        elif node.get("name"):
            return f'//*[@name="{node["name"]}"]'
        elif node.get("id"):   # harmonyos, ios
            return f'//*[@id="{node["id"]}"]'
        return None

    def _build_xpath(self, node: Dict, path: str, found_value: bool = False) -> str:
        """
        Recursively builds the XPath expression.

        Args:
        node (Dict): The current node.
        path (str): The current path.
        found_value (bool, optional): Whether a specific attribute value has been found. Default is False.

        Returns:
        str: The complete XPath expression.
        """
        if not node:
            return path
        value = self._get_value(node)
        if value:
            found_value = True
            return value + path
        if self.platform == 'ios' and not (node.get("lable") or node.get("name")):
            # If the platform is iOS and the node does not have lable, name, build from root
            return self._build_from_root(node, path)

        parent_node = self._find_node_by_id(node["_parentId"])
        if parent_node:
            siblings = parent_node.get("children", [])
            index = 1
            for sibling in siblings:
                if sibling["_type"] == node["_type"]:
                    if sibling["_id"] == node["_id"]:
                        break
                    index += 1
            path = f'/{node["_type"]}[{index}]' + path
            return self._build_xpath(parent_node, path, found_value)
        return path

    def _build_from_root(self, node: Dict, path: str) -> str:
        """
        Builds the XPath expression from the root node.

        Args:
        node (Dict): The current node.
        path (str): The current path.

        Returns:
        str: The complete XPath expression.
        """
        if "_type" in node:
            parent_node = self._find_node_by_id(node["_parentId"])
            if parent_node:
                siblings = parent_node.get("children", [])
                index = 1
                for sibling in siblings:
                    if sibling["_type"] == node["_type"]:
                        if sibling["_id"] == node["_id"]:
                            break
                        index += 1
                path = f'/{node["_type"]}[{index}]' + path
                return self._build_from_root(parent_node, path)
        return '//' + path.lstrip('/')

    def get_xpathLite(self, target_id: str) -> str:
        """
        Gets the XPathLite path for the target node.

        Args:
        target_id (str): The target node ID.

        Returns:
        str: The XPathLite path.
        """
        target_node = self._find_node_by_id(target_id)
        if not target_node:
            return None

        xpath_lite = self._build_xpath(target_node, "")
        if not xpath_lite.startswith('//*[@'):
            xpath_lite = self._build_from_root(target_node, "")

        return xpath_lite


# -*- coding: utf-8 -*-

import logging

formatter = logging.Formatter('[%(asctime)s] %(filename)15s[line:%(lineno)4d] \
                              [%(levelname)s] %(message)s',
                              datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger('hmdriver2')
logger.setLevel(logging.DEBUG)

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)

logger.addHandler(console_handler)


__all__ = ['logger']



# -*- coding: utf-8 -*-

import uuid
from typing import Dict


def convert_harmony_hierarchy(data: Dict) -> Dict:
    ret = {"_id": str(uuid.uuid4()), "children": [], "_parentId": ""}

    def __travel(node_a, parent_id=""):
        node_b = {
            "index": 0,
            "text": "",
            "id": "",
            "_type": "",
            "description": "",
            "checkable": False,
            "clickable": False,
            "enabled": False,
            "focusable": False,
            "focused": False,
            "scrollable": False,
            "longClickable": False,
            "password": False,
            "selected": False,
            "rect": {
                "x": 0,
                "y": 0,
                "width": 0,
                "height": 0
            },
            "_id": str(uuid.uuid4()),
            "_parentId": parent_id,
            "xpath": ""
        }

        attributes = node_a.get("attributes", {})
        node_b["xpath"] = attributes.get("xpath", "")
        node_b["_type"] = attributes.get("type", "")
        node_b["id"] = attributes.get("id", "")
        node_b["description"] = attributes.get("description", "")
        node_b["text"] = attributes.get("text", "")
        node_b["checkable"] = attributes.get("checkable", "").lower() == "true"
        node_b["clickable"] = attributes.get("clickable", "").lower() == "true"
        node_b["enabled"] = attributes.get("enabled", "").lower() == "true"
        node_b["focusable"] = attributes.get("focusable", "").lower() == "true"
        node_b["focused"] = attributes.get("focused", "").lower() == "true"
        node_b["scrollable"] = attributes.get("scrollable", "").lower() == "true"
        node_b["longClickable"] = attributes.get("longClickable", "").lower() == "true"
        bounds = attributes.get("bounds", "")
        bounds = bounds.strip("[]").split("][")
        node_b["rect"]["x"] = int(bounds[0].split(",")[0])
        node_b["rect"]["y"] = int(bounds[0].split(",")[1])
        node_b["rect"]["width"] = int(bounds[1].split(",")[0]) - int(bounds[0].split(",")[0])
        node_b["rect"]["height"] = int(bounds[1].split(",")[1]) - int(bounds[0].split(",")[1])

        children = node_a.get("children", [])
        if children:
            node_b["children"] = [__travel(child, node_b["_id"]) for child in children]

        return node_b

    # Recursively convert children of a to match b's structure
    ret["children"] = [__travel(child, ret["_id"]) for child in data.get("children", [])]

    return ret


# -*- coding: utf-8 -*-

from typing import Union, Dict, Any

from fastapi import APIRouter, Query
from fastapi.responses import RedirectResponse

from uiviewer._device import (
    list_serials,
    init_device,
    cached_devices,
    AndroidDevice,
    IosDevice,
    HarmonyDevice
)
from uiviewer._version import __version__
from uiviewer._models import ApiResponse, XPathLiteRequest
from uiviewer.parser.xpath_lite import XPathLiteGenerator


router = APIRouter()


@router.get("/")
def root():
    return RedirectResponse(url="/static/index.html")


@router.get("/health")
def health():
    return "ok"


@router.get("/version", response_model=ApiResponse)
def get_version():
    return ApiResponse.doSuccess(__version__)


@router.get("/{platform}/serials", response_model=ApiResponse)
def get_serials(platform: str):
    serials = list_serials(platform)
    return ApiResponse.doSuccess(serials)


@router.post("/{platform}/{serial}/connect", response_model=ApiResponse)
def connect(
    platform: str,
    serial: str,
    wdaUrl: Union[str, None] = Query(None),
    maxDepth: Union[int, None] = Query(None)
):
    ret = init_device(platform, serial, wdaUrl, maxDepth)
    return ApiResponse.doSuccess(ret)


@router.get("/{platform}/{serial}/screenshot", response_model=ApiResponse)
def screenshot(platform: str, serial: str):
    device: Union[AndroidDevice, IosDevice, HarmonyDevice] = cached_devices.get((platform, serial))
    data = device.take_screenshot()
    return ApiResponse.doSuccess(data)


@router.get("/{platform}/{serial}/hierarchy", response_model=ApiResponse)
def dump_hierarchy(platform: str, serial: str):
    device: Union[AndroidDevice, IosDevice, HarmonyDevice] = cached_devices.get((platform, serial))
    data = device.dump_hierarchy()
    return ApiResponse.doSuccess(data)


@router.post("/{platform}/hierarchy/xpathLite", response_model=ApiResponse)
async def fetch_xpathLite(platform: str, request: XPathLiteRequest):
    tree_data = request.tree_data
    node_id = request.node_id
    generator = XPathLiteGenerator(platform, tree_data)
    xpath = generator.get_xpathLite(node_id)
    return ApiResponse.doSuccess(xpath)


# -*- coding: utf-8 -*-

import argparse
from uiviewer.__main__ import run


def main():
    parser = argparse.ArgumentParser(description="My CLI Tool")
    parser.add_argument('-p', '--port', type=int, default=8000, help='local listen port for uiviewer')
    args = parser.parse_args()
    run(port=args.port)


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

import enum

from pydantic import BaseModel
from typing import Any, Union, Dict, Tuple, Optional


class Platform(str, enum.Enum):
    ANDROID = "android"
    IOS = "ios"
    HARMONY = "harmony"


class ApiResponse(BaseModel):
    success: bool = True
    data: Any = None
    message: Union[str] = None

    @classmethod
    def doSuccess(cls, data):
        return ApiResponse(success=True, data=data, message=None)

    @classmethod
    def doError(cls, message):
        return ApiResponse(success=False, data=None, message=message)


class BaseHierarchy(BaseModel):
    jsonHierarchy: Optional[Dict] = None
    windowSize: Tuple[int, int]
    scale: int = 1
    activityName: Optional[str] = None
    packageName: Optional[str] = None


class XPathLiteRequest(BaseModel):
    tree_data: Dict[str, Any]
    node_id: str


# -*- coding: utf-8 -*-

import xml.dom.minidom
import uuid
from typing import Dict

from uiviewer.parser.utils import parse_bounds, safe_xmlstr, str2bool, str2int, convstr

__alias = {
    'class': '_type',
    'resource-id': 'resourceId',
    'content-desc': 'description',
    'long-clickable': 'longClickable',
    'bounds': 'rect',
}

__parsers = {
    '_type': safe_xmlstr,  # node className
    # Android
    'rect': parse_bounds,
    'text': convstr,
    'resourceId': convstr,
    'package': convstr,
    'checkable': str2bool,
    'scrollable': str2bool,
    'focused': str2bool,
    'clickable': str2bool,
    'selected': str2bool,
    'longClickable': str2bool,
    'focusable': str2bool,
    'password': str2bool,
    'index': int,
    'description': convstr,
    # iOS
    'name': convstr,
    'label': convstr,
    'x': str2int,
    'y': str2int,
    'width': str2int,
    'height': str2int,
    # iOS && Android
    'enabled': str2bool,
}


def _parse_node_attributes(node):
    attributes = {}
    for key, value in node.attributes.items():
        key = __alias.get(key, key)
        parser = __parsers.get(key)
        if value is None:
            attributes[key] = None
        elif parser:
            attributes[key] = parser(value)
    return attributes


def _parse_uiautomator_node(node):
    attributes = _parse_node_attributes(node)
    if 'bounds' in attributes:
        lx, ly, rx, ry = map(int, attributes.pop('bounds'))
        attributes['rect'] = dict(x=lx, y=ly, width=rx - lx, height=ry - ly)
    return attributes


def convert_android_hierarchy(page_xml: str) -> Dict:
    dom = xml.dom.minidom.parseString(page_xml)
    root = dom.documentElement

    def __travel(node, parent_id=""):
        if node.attributes is None:
            return
        json_node = _parse_uiautomator_node(node)
        json_node['_id'] = str(uuid.uuid4())
        json_node['_parentId'] = parent_id
        json_node['xpath'] = ""
        json_node.pop("package", None)
        if node.childNodes:
            children = []
            for n in node.childNodes:
                child = __travel(n, json_node['_id'])
                if child:
                    children.append(child)
            json_node['children'] = children

        # Sort the keys
        keys_order = ['xpath', '_type', 'resourceId', 'text', 'description']
        sorted_node = {k: json_node[k] for k in keys_order if k in json_node}
        sorted_node.update({k: json_node[k] for k in json_node if k not in keys_order})

        return sorted_node

    return __travel(root)


# -*- coding: utf-8 -*-

import re


def parse_bounds(text):
    m = re.match(r'\[(\d+),(\d+)\]\[(\d+),(\d+)\]', text)
    if m is None:
        return None
    (lx, ly, rx, ry) = map(int, m.groups())
    return dict(x=lx, y=ly, width=rx - lx, height=ry - ly)


def safe_xmlstr(s):
    return s.replace("$", "-")


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def str2int(v):
    return int(v)


def convstr(v):
    return v


# -*- coding: utf-8 -*-

import uuid
from typing import Dict


def convert_ios_hierarchy(data: Dict, scale: int) -> Dict:

    def __travel(node, parent_id=""):
        node['_id'] = str(uuid.uuid4())
        node['_parentId'] = parent_id
        node['_type'] = node.pop('type', "null")
        node['id'] = node.pop('rawIdentifier', "null")
        node['xpath'] = ""
        if 'rect' in node:
            rect = node['rect']
            node['rect'] = {k: v * scale for k, v in rect.items()}

        # Recursively process children nodes
        if 'children' in node:
            node['children'] = [__travel(child, node['_id']) for child in node['children']]

        # Sort the keys
        keys_order = ['xpath', '_type', 'label', 'name', 'id', 'value']
        sorted_node = {k: node[k] for k in keys_order if k in node}
        sorted_node.update({k: node[k] for k in node if k not in keys_order})

        return sorted_node

    return __travel(data)
